// ADSmartIO
//
//   Select the platform that most closely matches your board.
//   (e.g '#define BITSY 1' for Bitsy, Bitsy+, BitsyX)
//
//   Uncomment the desired tests in main().
//
//   Compile like this:
//     arm-linux-g++ -O2 -I. -L. sio.cc -lsio -o siotest
//
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "libsio.h"

#define GETBIT(buf,bit) (((buf) & (1 << (bit))) ? 1 : 0)
#define PRINTBIT(mask,buf,bit) (GETBIT((mask),(bit)) ? printf("%d", GETBIT((buf),(bit))) : printf("-"))

// Local prototypes
void print_device_info(void);
void test_back_light(void);
void test_keypad(int rows, int cols);
void test_adc(void);
void test_adc_temp(void);
void test_dio(int dir, int val);
void test_ucb_dio(void);
void test_ucb_din(void);
void test_ucb_adc(void);
void test_eeprom_read(int first, int last);
void test_eeprom_write(int dest);

// These are mutually exclusive
#define BITSY 1
#define GMSTR 0
#define ADVGC 0
#define GCPLS 0

#define DIGOPT 3 // should be 2 for older AVR firmware

int main(int argc, char *argv[])
{
  print_device_info();

  //test_back_light();
  //test_keypad((argc > 1) ? strtoul(argv[1], NULL, 10) : 0, (argc > 2) ? strtoul(argv[2], NULL, 10) : 0 );
  //test_adc();
  //test_adc_temp();
  //test_dio((argc > 1) ? strtoul(argv[1], NULL, 0x10) : 0, (argc > 2) ? strtoul(argv[2], NULL, 0x10) : 0 );
  //test_ucb_dio();
  //test_ucb_din();
  //test_ucb_adc();
  //test_eeprom_read(0, 15);
  //test_eeprom_write(10);

  return 0;
}

void print_device_info(void)
{
  printf("Device Version  : %c\n", SIOReadDeviceVersion());
  printf("Device Part Type: %x\n", SIOReadDevicePartType());
  printf("Firmware Version: %x\n", SIOReadFirmwareVersion());
  sleep(1);
}

void test_back_light(void)
{
  int i;

  SIOBackLightOff();
  sleep(1);
  SIOBackLightOn();
  for (i = 0xff; i >= 0; i--) {
    SIOSetBackLightPwm(i);
  }
}

// Digital IO, Keypad, & A/D ---------------------------------------

#if BITSY
#  define SIO_DIG_A 0x0f
#  define SIO_DIG_B 0x00
#  define SIO_DIG_C 0x1f // 0x0f // 0xdf - pin 5 is 32 kHz crystal
#  define SIO_DIG_D 0x00 // 0x03 - panel & temperature stuff
#  define MAX_A_PORTS 4
#  define MAX_B_PORTS 0
#  define MAX_C_PORTS 5 // 4 // 8 - really 7 but the hole is in the middle
#  define MAX_D_PORTS 0 // 2
#  define ADC_TEMP_PORT 4
#elif ADVGC
#  define SIO_DIG_A 0xff
#  define SIO_DIG_B 0x00 // 0x02
#  define SIO_DIG_C 0x1f // pin 7 & 6 are RTC, 5 may be 32 kHz crystal
#  define SIO_DIG_D 0xf2 // pin 3 is RTC, 2 is power enable, 0 is ?
#  define MAX_A_PORTS 8
#  define MAX_B_PORTS 0 // 2
#  define MAX_C_PORTS 5
#  define MAX_D_PORTS 8
#else
#  define SIO_DIG_A 0xff
#  define SIO_DIG_B 0x0f
#  define SIO_DIG_C 0xff
#  define MAX_A_PORTS 8
#  define MAX_B_PORTS 4
#  define MAX_C_PORTS 8
#  if GMSTR
#    define SIO_DIG_D 0x7f
#    define MAX_D_PORTS 7
//#    define ADC_TEMP_PORT 4
#  else // GCPLS
#    define SIO_DIG_D 0x03
#    define MAX_D_PORTS 2
#  endif
#endif

void test_dio(int dir, int val)
{
  int buf, mask, i, j, k = dir ? 2 : 1;

  // Added for Bitsy+ AVR which doesn't use SelectOption to change modes
  SIOSetKeypadSize(0,0);

  // Option 3 is all digital I/O
  // Option 2 is A/D on A and digital I/O on B, C, & D
  // Option 1 is all digital I/O on unused keypad pins
  // Option 3 is NOT supported by all AVR versions (see also write_sio_select())
  // Option 1 with a 0 x 0 keypad is indeterminate
  SIOSelectOption(DIGOPT);

// Set direction of digital I/Os
// Turn on pull up resistors by writing 1s to inputs

#if (DIGOPT != 2)
  SIOInitDDRA(dir & SIO_DIG_A);
  SIOWritePortA(val & SIO_DIG_A);
#endif

#if MAX_B_PORTS
  SIOInitDDRB(dir & SIO_DIG_B);
  SIOWritePortB(val & SIO_DIG_B);
#endif

  SIOInitDDRC(dir & SIO_DIG_C);
  SIOWritePortC(val & SIO_DIG_C);

#if MAX_D_PORTS
  SIOInitDDRD(dir & SIO_DIG_D);
  SIOWritePortD(val & SIO_DIG_D);
#endif

  printf("\n");
  for (j=0; j < 60; j += k) {
    //    if (j % 22 == 0) {
    if (j % 10 == 0) {
#if (DIGOPT == 2)
#  if BITSY
      printf("C\n");
      printf("43210\n");
#  elif ADVGC
//     printf("B  C     D\n");
//     printf("1- 43210 7654--1-\n");
      printf("C     D\n");
      printf("43210 765?--?-\n");
#  elif GMSTR
      printf("B    C        D\n");
      printf("3210 76543210 6543210\n");
#  else
      printf("B    C        D\n");
      printf("3210 76543210 10\n");
#  endif
#else
#  if BITSY
      printf("A    C\n");
      printf("3210 43210\n");
#  elif ADVGC
//     printf("A        B  C     D\n");
//     printf("76543210 1- 43210 7654--1-\n");
      printf("A        C     D\n");
      printf("76543210 43210 765?--?-\n");
#  elif GMSTR
      printf("A        B    C        D\n");
      printf("76543210 3210 76543210 6543210\n");
#  else // GCPLS
      printf("A        B    C        D\n");
      printf("76543210 3210 76543210 10\n");
#  endif
#endif
    }

#if (DIGOPT != 2)
    buf  = SIOReadPortA();
    mask = SIO_DIG_A|~dir;
    SIOWritePortA((buf^dir) & SIO_DIG_A); // toggle outputs / sample & hold inputs
    for (i=MAX_A_PORTS-1; i>=0; i--)
      PRINTBIT(mask,buf,i);

    printf(" ");
#endif

#if MAX_B_PORTS
    buf  = SIOReadPortB();
    mask = SIO_DIG_B|~dir;
    SIOWritePortB((buf^dir) & SIO_DIG_B); // toggle outputs / sample & hold inputs
    for (i=MAX_B_PORTS-1; i>=0; i--)
      PRINTBIT(mask,buf,i);

    printf(" ");
#endif

    buf  = SIOReadPortC();
    mask = SIO_DIG_C|~dir;
    SIOWritePortC((buf^dir) & SIO_DIG_C); // toggle outputs / sample & hold inputs
    for (i=MAX_C_PORTS-1; i>=0; i--)
      PRINTBIT(mask,buf,i);

#if MAX_D_PORTS
    printf(" ");

    buf  = SIOReadPortD();
    mask = SIO_DIG_D|~dir;
    SIOWritePortD((buf^dir) & SIO_DIG_D); // toggle outputs / sample & hold inputs
    for (i=MAX_D_PORTS-1; i>=0; i--)
      PRINTBIT(mask,buf,i);

#endif

    printf("\n");
    sleep(k);
  }
}

#if BITSY
#define  MAX_ROWS MAX_C_PORTS
#define  MAX_COLS MAX_A_PORTS
#define TEST_ROWS 2 // MAX 5
#define TEST_COLS 3 // MAX 4
#elif ADVGC
#define  MAX_ROWS 5 // 8
#define  MAX_COLS 5 // 8
#define TEST_ROWS 4 // MAX 8
#define TEST_COLS 4 // MAX 8
#else
#define  MAX_ROWS MAX_C_PORTS
#define  MAX_COLS MAX_A_PORTS
#define TEST_ROWS 6 // MAX 8
#define TEST_COLS 3 // MAX 8
#endif

void test_keypad(int rows, int cols)
{
  int buf, ok = 4;
  unsigned char portc = SIO_DIG_C;

  if (rows < 1 || rows > MAX_ROWS)
    rows = TEST_ROWS;
  if (cols < 1 || cols > MAX_COLS)
    cols = TEST_COLS;

  SIOInitDDRA(SIO_DIG_A);   // port A (cols) are outputs
  SIOWritePortA(0);         // set initial outputs low
  SIOInitDDRC(0);           // port C (rows) are inputs
  SIOWritePortC(portc);     // turn input pull ups on
#if 0
  SIOInitDDRD(0x05);        // port D (both) mixed I/O
  SIOWritePortD(0xe0);      // turn input pull ups on
#endif
  SIOSelectOption(1);       // keypad mode - (unused rows/cols are digital I/O)

  printf("Configure keypad (%d rows, %d cols)\n", rows, cols);
  SIOSetKeypadSize(cols, rows);

#if 1
  if (rows < MAX_C_PORTS && (SIO_DIG_C & (1<<rows)) == (1<<rows)) {
    printf("\nVerify digital I/O on unused pin C%d\n", rows);
    SIOInitDDRC(1<<rows);   // pin Cx output (first non-row pin)
    // check initial state and do at least one write
    buf = SIOReadPortC();
    printf("  Reading %02x on Port C - C%d is %d\n", buf, rows, GETBIT(buf,rows));
    portc ^= (1<<rows);     // toggle Cx
    SIOWritePortC(portc);   // write Cx output 0
    printf("  Writing %02x on Port C\n", portc);
    // read back Cx verify output 0
    buf = SIOReadPortC();
    printf("  Reading %02x on Port C - C%d is %d\n", buf, rows, GETBIT(buf,rows));
    if (GETBIT(buf,rows) == 1) {
      printf("Digital I/O on unused pin C%d failed write/read 0\n", rows);
    }
    else {
      // now write again to verify value changed
      portc ^= (1<<rows);   // toggle Cx
      SIOWritePortC(portc); // write Cx output 1
      printf("  Writing %02x on Port C\n", portc);
      // read back Cx verify output 1
      buf = SIOReadPortC();
      printf("  Reading %02x on Port C - C%d is %d\n", buf, rows, GETBIT(buf,rows));
      if (GETBIT(buf,rows) == 0)
        printf("Digital I/O on unused pin C%d failed write/read 1\n", rows);
      else
	printf("Digital I/O on unused pin C%d verified\n", rows);
    }
  }
#endif

#if 0
  printf("\nConfigure keypad with 1 sec timeout\n\n");
  SetKeypadTimeout(1000);
  {
    int j;
    for (j=0; j < 60; j++) {
      buf = SIOReadKeypad();
      if (0 > buf) {
        printf("Keypad = %d (timed out - no key)\n", buf);
      }
      else {
        buf &= 0x3f;
        printf("Keypad %d = col %d, row %d\n", buf, buf/rows, buf%rows);
        if (0 == buf) {
          printf("Got 0 - exiting loop\n");
          break;
	}
        if (rows*cols <= buf) {
          printf("Invalid key!\nExiting loop.\n");
	  break;
	}
      }
    }
  }
#endif

  printf("\nConfigure keypad to wait for keypresses\n\n");
  SetKeypadTimeout(-1);
  while (ok) {
    buf = SIOReadKeypad() & 0x3f;
    printf("Keypad %d = col %d, row %d\n", buf, buf/rows, buf%rows);
    if (rows*cols <= buf) {
      --ok;
      printf("Invalid key!\n");
    }
  }
  printf("Giving up!\n");

  // Added for Bitsy+ AVR which doesn't use SelectOption to change modes
  SIOSetKeypadSize(0,0);
}

void test_adc(void)
{
  int buf, ch, i;

  SIOSetKeypadSize(0,0);
  SIOInitDDRA(0);
  SIOWritePortA(0);
#ifdef ADC_TEMP_PORT
  SIOThermistorOff();
#endif

  SIOSelectOption(2);

  printf("Read A/D converters\n\n");
  while (1) {
    sleep(1);
    for (i=0; i<MAX_A_PORTS; i++) {
      ch = i;
      buf = SIOAnalogConversion(ch);
      printf("ch%d = %x\n", i, buf);
    }
    printf("\n");
  }
}

#ifdef ADC_TEMP_PORT
void test_adc_temp(void)
{
  int buf, ch, i;

  SIOSetKeypadSize(0,0);
  SIOInitDDRA(0);
  SIOWritePortA(0);

  SIOSelectOption(2);
  SIOThermistorOn();

  printf("Read Thermistor A/D\n\n");
  while (1) {
    sleep(1);
    buf = SIOAnalogConversion(ADC_TEMP_PORT);
    printf("temp = %x\n", buf);
  }

  SIOThermistorOff();
}
#else
void test_adc_temp(void)
{
  printf("Thermistor A/D Port not specified.\n\n");
}
#endif

// UCB1200 ---------------------------------------------

// For boards without a UCB1200, the digital I/O
// may be provided by a CPLD or GPIO substitute.
// There is no substitute for the UCB1200 A/D.

void test_ucb_dio(void)
{
  unsigned short buf;
  int i, j;

  SIOInitUCB(0x3ff);
  printf("Write 'UCB' digital I/O\n\n");
  printf("9876543210\n");
  sleep(2);
  for (j=0; j < 10; j++) {
    buf = 1<<j;
    SIOWriteUCB(buf);
    buf = SIOReadUCBDigital();
    if ((unsigned)~0 == buf) {
      printf("----------\n");
      continue;
    }
    for (i=9; i>=0; i--)
      printf("%d", GETBIT(buf,i));
    printf("\n");
    sleep(5);
  }

  SIOWriteUCB(0);
  SIOInitUCB(0);
  //  SIOWriteUCB(0);
  printf("\nRead 'UCB' digital I/O\n\n");
  for (j=0; j < 120; j++) {
    if (j % 20 == 0) {
      printf("\n");
      sleep(2);
      SIOInitUCB(0x3ff);
      SIOWriteUCB(0x3ff);
      printf("9876543210\n");
    }
    SIOInitUCB(0);
    sleep(1);
    buf = SIOReadUCBDigital();
    if ((unsigned)~0 == buf) {
      printf("----------\n");
      continue;
    }
    for (i=9; i>=0; i--)
      printf("%d", GETBIT(buf,i));
    printf("\n");
  }
}

void test_ucb_din(void)
{
  unsigned short buf;
  int i, j;

  SIOInitUCB(0);
  //  SIOWriteUCB(0);
  printf("\nRead 'UCB' digital I/O\n\n");
  for (j=0; j < 120; j++) {
    if (j % 22 == 0) {
      printf("\n");
      printf("9876543210\n");
    }
    sleep(1);
    buf = SIOReadUCBDigital();
    if ((unsigned)~0 == buf) {
      printf("----------\n");
      continue;
    }
    for (i=9; i>=0; i--)
      printf("%d", GETBIT(buf,i));
    printf("\n");
  }
}

void test_ucb_adc(void)
{
  unsigned short buf, i;

  printf("Read UCB 1200 A/D converters\n\n");
  while (1) {
    sleep(1);
    for (i=0; i<4; i++) {
      buf = SIOReadUCBAnalog(i);
      if ((unsigned)~0 != buf)
        printf("ch%d = %d\n", i, buf);
    }
    printf("\n");
  }
}

// EEPROM ----------------------------------------------

// Warning:  EEPROM may not be supported by driver/firmware.

#include <fcntl.h>
#define BANK_SIZE 16
#define MAX_BANKS 16 // 32

long read_eeprom(unsigned char *buf, unsigned char bank)
{
  int fd = open("/dev/sio13", O_RDWR);
  if (fd < 0)
  {
    printf("Cannot open /dev/sio13\n");
    return (long)fd;
  }

  long pos = lseek(fd, bank*BANK_SIZE, SEEK_SET); // SEEK_END not defined
  //printf("lseek(%d, %d, %d) : %ld\n", fd, bank*BANK_SIZE, SEEK_SET, pos);
  if (pos != bank*BANK_SIZE)
  {
    printf("Cannot find offset %d\n", bank*BANK_SIZE);
    close(fd);
    return pos;
  }

  long rtn = read(fd, buf, BANK_SIZE);
  close(fd);
  return rtn;
}

long write_eeprom(unsigned char *buf, unsigned char bank)
{
  int fd = open("/dev/sio13", O_RDWR);
  if (fd < 0)
  {
    printf("Cannot open /dev/sio13\n");
    return (long)fd;
  }

  long pos = lseek(fd, bank*BANK_SIZE, SEEK_SET); // SEEK_END not defined
  //printf("lseek(%d, %d, %d) : %ld\n", fd, bank*BANK_SIZE, SEEK_SET, pos);
  if (pos != bank*BANK_SIZE)
  {
    printf("Cannot find offset %d\n", bank*BANK_SIZE);
    close(fd);
    return pos;
  }

  long rtn = write(fd, buf, BANK_SIZE);
  close(fd);
  return rtn;
}

void test_eeprom_read(int first, int last)
{
  unsigned char bank[4*BANK_SIZE];
  long res;
  int i, j;

  if (0 > first)
    first = 0;
  if (MAX_BANKS <= last)
    last = MAX_BANKS - 1;

  for (i = first; i <= last; i++) {
    res = read_eeprom(bank, i);
    if (BANK_SIZE != res) {
      printf("Bank %02d: %x\n", i, (int)res);
      break;
    }
    printf("Bank %02d:", i);
    for (j = 0; j < BANK_SIZE; j++) {
      printf(" %02x", bank[j]);
    }
    printf("\n");
  }
  printf("\n");
  sleep(1);
}

void test_eeprom_write(int dest)
{
  unsigned char bank[4*BANK_SIZE];
  long res;
  int j;

  if (0 > dest || MAX_BANKS <= dest)
    return;

  bank[0] = 0xAA;
  bank[1] = 0x55;
  res = write_eeprom(bank, dest);
  sleep(1);

  if (BANK_SIZE != res) {
    printf("Write Bank %02d: %x\n\n", dest, (int)res);
    return;
  }

  printf("Wrote  :");
  for (j = 0; j < 16; j++) {
    printf(" %02x", bank[j]);
  }
  printf(" to bank %02d\n\n", dest);
  sleep(1);

  test_eeprom_read(dest - 1, dest + 1); // read clips if out of range
}

// ADSmartIO
//  - listed by port
//
//            GC+ & GM        Bitsy
//           -----------   -----------
// Dig A/D   Key  Jumper   Key  Jumper
// A0  ch0   Col0 J7  17   Col0 J10 27
// A1  ch1   Col1 J7  19   Col1 J10 29
// A2  ch2   Col2 J7  21   Col2 J10 31
// A3  ch3   Col3 J7  23   Col3 J10 33
// A4  ch4   Col4 J7  25
// A5  ch5   Col5 J7  27
// A6  ch6   Col6 J7  29
// A7  ch7   Col7 J7  31
// B0             J2  12
// B1             J2  14
// B2             J2  16
// B3             J2  18
// C0        Row7 J7  15   Row0 J10 17
// C1        Row6 J7  13   Row1 J10 19
// C2        Row5 J7  11   Row2 J10 21
// C3        Row4 J7   9   Row3 J10 23
// C4        Row3 J7   7   Row4 J10 25
// C5        Row2 J7   5
// C6        Row1 J7   3        J3  35
// C7        Row0 J7   1        J3  37
// D0             J2  20        J10 15
// D1             J2  22        J10 13
//
// ADSmartIO
//  - listed by jumper
//
//       GC+ & GM                Bitsy
// --------------------   --------------------
// Jumper  Dig A/D Key    Jumper  Dig A/D Key
// J2  12  B0             J3  35  C6
//     14  B1                 37  C7
//     16  B2             J10 13  D1
//     18  B3                 15  D0
//     20  D0                 17  C0      Row0
//     22  D1                 19  C1      Row1
// J7   1  C7      Row0       21  C2      Row2
//      3  C6      Row1       23  C3      Row3
//      5  C5      Row2       25  C4      Row4
//      7  C4      Row3       27  A0  ch0 Col0
//      9  C3      Row4       29  A1  ch1 Col1
//     11  C2      Row5       31  A2  ch2 Col2
//     13  C1      Row6       33  A3  ch3 Col3
//     15  C0      Row7
//     17  A0  ch0 Col0
//     19  A1  ch1 Col1
//     21  A2  ch2 Col2
//     23  A3  ch3 Col3
//     25  A4  ch4 Col4
//     27  A5  ch5 Col5
//     29  A6  ch6 Col6
//     31  A7  ch7 Col7
